"""
gbr2laser V0.11alpha
Calculation of gcode for pcb-exposure with a UV-laser on a CNC-machine.
Datasource: gbr-export from KiCAD 4.0.2 stable
Datatarget: G-Code for LinuxCNC
Tested in python 2.7.6
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <FlorenzW> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return. Florian W.
 * ----------------------------------------------------------------------------
"""

import math
from shapely.geometry import *
from shapely.affinity import *
from shapely.ops import cascaded_union
import matplotlib.pyplot as plt
from copy import deepcopy
import fpformat
from random import randint

ax = []

xMin = 0
yMin = 0
xMax = 10
yMax = 10
import_as_perimeter = False


def optimizing_outline_arrangement_V2(perimeter):
    all_polylines = []
    i = 0
    while i < len(perimeter):
        j = 0
        polyline = []
        while j < len(perimeter[i].coords):
            polyline.append(perimeter[i].coords[j])
            j = j + 1
        all_polylines.append(polyline)
        i = i + 1

    print("")    
    print("Optimizing outline arrangement, this takes a while...")
    print("Positioninglength old: "),
    print(calculate_positioning_length_outlines(all_polylines))

    all_polylines_new = []
    punkt_old = [0, 0]

    while len(all_polylines) > 0:
        i = 0
        length_old = line_length([punkt_old, all_polylines[i][0]])
        linenumber = 0
        while i < len(all_polylines):
            length_new = line_length([punkt_old, all_polylines[i][0]])
            if length_new < length_old:
                length_old = length_new
                linenumber = i
            i = i + 1

        punkt_old = deepcopy(all_polylines[linenumber][0])
        all_polylines_new.append(all_polylines[linenumber])
        del all_polylines[linenumber]

    print("")      
    print("Positioninglength new: "),
    print(calculate_positioning_length_outlines(all_polylines_new))

    perimeter = []
    i = 0
    while i < len(all_polylines_new):
        perimeter.append(all_polylines_new[i])
        i = i + 1

    return(MultiLineString(perimeter))
            


def optimizing_outline_arrangement(perimeter):
    global optimizing_cycles
    all_polylines = []
    i = 0
    while i < len(perimeter):
        j = 0
        polyline = []
        while j < len(perimeter[i].coords):
            polyline.append(perimeter[i].coords[j])
            j = j + 1
        all_polylines.append(polyline)
        i = i + 1

    print("")    
    print("Optimizing outline arrangement, this takes a while...")
    print("Positioninglength old: "),
    print(calculate_positioning_length_outlines(all_polylines))

    t = 0
    while t < optimizing_cycles:

        i = 0
        while i < 1000:

            length_old = calculate_positioning_length_outlines(all_polylines)
            random_posA = randint(0,len(all_polylines)-1)
            random_posB = randint(0,len(all_polylines)-1)
            all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]
            if length_old <= calculate_positioning_length_outlines(all_polylines): 
                all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]
            i = i + 1

        t = t + 1
        print("Cycle: "),
        print(t),
        print("  "),
        print(length_old)


    print("")      
    print("Positioninglength new: "),
    print(calculate_positioning_length_outlines(all_polylines))

    perimeter = []
    i = 0
    while i < len(all_polylines):
        perimeter.append(all_polylines[i])
        i = i + 1

    return(MultiLineString(perimeter))


def calculate_positioning_length_outlines(all_polylines):
    i = 0
    length = 0
    while i < len(all_polylines)-1:
        length = length + line_length([all_polylines[i][0],all_polylines[i+1][0]])
        i = i + 1
    return length


#Returns the length of a line
def line_length(line):
    return math.sqrt(pow(line[0][0]-line[1][0],2) + pow(line[0][1]-line[1][1],2))


def export_gcode(geometriy_outlines, filling_pattern):
    print("")
    print("Exporting GCode")
    DateiOutput = open(filename_output,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G64 P0.001")
    DateiOutput.write("\n")
#    DateiOutput.write("M5")
    DateiOutput.write("M63 P-0")
    DateiOutput.write("\n")
    DateiOutput.write("F")
    DateiOutput.write(str(speed_perimeters))
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0 Z0.0")
    DateiOutput.write("\n")

    i = 0
    while (i < len(geometriy_outlines)):

        line = LineString(geometriy_outlines[i])
        x, y = line.coords.xy

        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(x[0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(y[0], 3))
        DateiOutput.write("\n")
#        DateiOutput.write("M3")
        DateiOutput.write("M62 P-0")
        DateiOutput.write("\n")

        j = 1
        while (j < len(geometriy_outlines[i].coords)-1):
            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(x[j], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(y[j], 3))
            DateiOutput.write("\n")
            j = j + 1

        DateiOutput.write("G1 X")
        DateiOutput.write(fpformat.fix(x[j], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(y[j], 3))
        DateiOutput.write("\n")
#        DateiOutput.write("M5")
        DateiOutput.write("M63 P-0")
        DateiOutput.write("\n")

        i = i + 1    
            
    DateiOutput.write("F"),
    DateiOutput.write(str(speed_filling))
    DateiOutput.write("\n")

    i = 0
    while i < len(filling_pattern):
        DateiOutput.write("G1 X")
        DateiOutput.write(fpformat.fix(filling_pattern[i][0][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(filling_pattern[i][0][1], 3))
        DateiOutput.write("\n")
#        DateiOutput.write("M3")
        DateiOutput.write("M62 P-0")
        DateiOutput.write("\n")
        DateiOutput.write("G1 X")
        DateiOutput.write(fpformat.fix(filling_pattern[i][1][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(filling_pattern[i][1][1], 3))
        DateiOutput.write("\n")
#        DateiOutput.write("M5")
        DateiOutput.write("M63 P-0")
        DateiOutput.write("\n")

        i = i + 1

    DateiOutput.write("G0 X0 Y0")
    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()


def plot_fillingpattern(filling_pattern, color):
    i = 0
    while i < len(filling_pattern):
        line = []
        line.append(filling_pattern[i][0])
        line.append(filling_pattern[i][1])
        plot_line(line, color)
        i = i + 1


def calculate_filling_pattern(all_lines):
    print("")
    print("Calculating fillingpattern - this takes a while...")
    global xMin
    global xMax
    global yMin
    global yMax
    global distance_fillinglines
    x = xMin
    fillingpattern = []
    filling_orientation = True   #is switched every line to turn the orientation

    while x < xMax:
        i = 0
        intersections = []
        while i < len(all_lines):

            pointK = []
            pointK.append(x)
            pointK.append(0)
            pointL = []
            pointL.append(x)
            pointL.append(1)
            pointM = []
            pointM = all_lines[i][0]
            pointN = []
            pointN = all_lines[i][1]
            
            if ((pointM[0] < x) and (pointN[0] > x)) or ((pointM[0] > x) and (pointN[0] < x)):
                pointS = calculate_intersection(pointK, pointL, pointM, pointN)
                intersections.append(pointS[1])

            i = i + 1

        if filling_orientation:
            intersections.sort(reverse = False)

            line = []
            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            fillingpattern.append(line)

        else:
            intersections.sort(reverse = True)

            line = []
            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            fillingpattern.append(line)
            
        x = x + distance_fillinglines

        filling_orientation = not filling_orientation  #Richtungswechsel bei jeder Zeile -> Wegoptimierung

    return fillingpattern


def explode_to_lines(all_polylines):
    i = 0
    lines = []
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i].coords)-1:
            line = []
            line.append(all_polylines[i].coords[j])
            if j < len(all_polylines[i].coords)-1:
                line.append(all_polylines[i].coords[j+1])
            else:
                line.append(all_polylines[i].coords[0])
            lines.append(line)
            j = j + 1
        i = i + 1
    return lines


def eliminate_vertical_lines(all_lines):
    i = 0
    lines = deepcopy(all_lines)
    while i < len(lines):
        if lines[i][0][0] == lines[i][1][0]:
            del(lines[i])
            i = -1
        i = i + 1
    return lines


def plot_geometry(geometry, color):
    i = 0
    while i < len(geometry):
        j = 0
        x = []
        y = []
        while j < len(geometry[i].coords):
            x.append(geometry[i].coords[j][0])
            y.append(geometry[i].coords[j][1])

            j = j + 1
        plot_lines(x, y, color)
        print(i),
        i = i + 1
    print("")


def plot_positioning(geometry, color):
    i = 0
    x = []
    y = []
    while i < len(geometry)-1:
        x.append(geometry[i].coords[0][0])
        y.append(geometry[i].coords[0][1])

        i = i + 1

    plot_lines(x, y, color)


# Tested only with KiCAD 4.0.2 gbr-export. No general import function.
def gbr_import(fname):
    global import_as_perimeter
    file_length = file_len(fname)
    line = ""
    tools = []
    geometry = []
    x_pos_old = 0
    y_pos_old = 0

    print("")
    print("Import tools"),
    print(filename_input)

    file_input = open(fname)

    i = 0
    while ((i < file_length) and (line <> "G04 APERTURE LIST")):
        line = file_input.readline()
        line = line.strip()   
        line = line.split("*")[0]
        i = i + 1

# -----------Import Tools------------
    while ((i < file_length) and (line <> "G04 APERTURE END LIST")):
        line = file_input.readline()
        line = line.strip()
        line = line.split("*")[0]

        if line.find(",", 0, len(line)) == 7:
            tool=[]
#           Toolname:
            tool.append(line[3:6])
#           Toolshape:
            tool.append(line[6:7])
#           Toolsize:
            line = line.split(",")[1]
            if line.find("X", 0, len(line)) == -1:
                tool.append(float(line))
            else:
                tool.append(float(line.split("X")[0]))
                tool.append(float(line.split("X")[1]))

                print(tool)
                
            tools.append(tool)
                    
        i = i + 1

# -----------Import geometry------------
    if len(tools) == 0:
        print("No tools found!")
    else:
        while ((i < file_length) and (line <> "M02*")):
            line = file_input.readline()
            line = line.strip()
            line = line.split("*")[0]

            if line[0:1] == "D":
                toolname = line[0:3]
                
                j = 0
                while j < len(tools):
                    if toolname == tools[j][0]:
                        tool = tools[j]
                    j = j + 1

            if line[0:1] == "X":
                x_string = line.split("-")[0]
                x_string = ''.join(i for i in x_string if i.isdigit())
                x_float = float(x_string) / 1000000
                
                y_string = line.split("-")[1]
                y_string = y_string.split("D")[0]
                y_string = ''.join(i for i in y_string if i.isdigit())
                y_float = float(y_string) / 1000000 * (-1)

                action_string = line.split("-")[1]
                action_string = action_string.split("D")[1]

#                print("X: "),
#                print(x_float),
#                print(" Y: "),
#                print(y_float),
#                print(" Action: "),
#                print(action_string)

                if action_string == "03":
#                    print(tool)
                    if import_as_perimeter:
                        if tool[1] == "R":
                            polygon = Polygon([(x_float -(tool[2]/2 + laser_diameter/2), y_float -(tool[3]/2 + laser_diameter/2)), (x_float -(tool[2]/2 + laser_diameter/2), y_float +(tool[3]/2 + laser_diameter/2)), (x_float +(tool[2]/2 + laser_diameter/2), y_float +(tool[3]/2 + laser_diameter/2)), (x_float +(tool[2]/2 + laser_diameter/2), y_float -(tool[3]/2 + laser_diameter/2)), (x_float -(tool[2]/2 + laser_diameter/2), y_float -(tool[3]/2 + laser_diameter/2))])                    
                        if tool[1] == "C":
                            polygon = Point(x_float, y_float).buffer(tool[2]/2 + laser_diameter/2)
                        if tool[1] == "O":
                            polygon = oval(x_float, y_float, tool[2]/2 + laser_diameter/2, tool[3]/2 + laser_diameter/2)
                    else:
                        if tool[1] == "R":
                            polygon = Polygon([(x_float -(tool[2]/2), y_float -(tool[3]/2)), (x_float -(tool[2]/2), y_float +(tool[3]/2)), (x_float +(tool[2]/2), y_float +(tool[3]/2)), (x_float +(tool[2]/2), y_float -(tool[3]/2)), (x_float -(tool[2]/2), y_float -(tool[3]/2))])                    
                        if tool[1] == "C":
                            polygon = Point(x_float, y_float).buffer(tool[2]/2)
                        if tool[1] == "O":
                            polygon = oval(x_float, y_float, tool[2]/2, tool[3]/2)
                    geometry.append(polygon)

                if action_string == "02":
#                    print(tool)
                    x_pos_old = x_float
                    y_pos_old = y_float

                if action_string == "01":
#                    print(tool)
                    if import_as_perimeter:
                        polygon = LineString([(x_pos_old, y_pos_old), (x_float, y_float)]).buffer(tool[2]/2 + laser_diameter/2) 
                    else:
                        polygon = LineString([(x_pos_old, y_pos_old), (x_float, y_float)]).buffer(tool[2]/2) 
                    x_pos_old = x_float
                    y_pos_old = y_float
                    geometry.append(polygon)
                
            i = i + 1
            
    return MultiPolygon(geometry)


def oval(x_pos, y_pos, x_radius, y_radius):
    polygon = []
    if (x_radius > y_radius):
        polygon = Polygon([(x_pos - (x_radius - y_radius), y_pos - y_radius), (x_pos + (x_radius - y_radius), y_pos - y_radius), (x_pos + (x_radius - y_radius), y_pos + y_radius), (x_pos - (x_radius - y_radius), y_pos + y_radius), (x_pos - (x_radius - y_radius), y_pos - y_radius)])
        polygon = polygon.union(Point(x_pos - (x_radius - y_radius), y_pos).buffer(y_radius))
        polygon = polygon.union(Point(x_pos + (x_radius - y_radius), y_pos).buffer(y_radius))
    else:
        polygon = Polygon([(x_pos - x_radius, y_pos - (y_radius - x_radius)), (x_pos + x_radius, y_pos - (y_radius - x_radius)), (x_pos + x_radius, y_pos + (y_radius - x_radius)), (x_pos - x_radius, y_pos + (y_radius - x_radius)), (x_pos - x_radius, y_pos - (y_radius - x_radius))])
        polygon = polygon.union(Point(x_pos, y_pos - (y_radius - x_radius)).buffer(x_radius))
        polygon = polygon.union(Point(x_pos, y_pos + (y_radius - x_radius)).buffer(x_radius))

    return polygon


def plot_line(line, color):
    global ax
    x = [line[0][0], line[1][0]]
    y = [line[0][1], line[1][1]]
    ax.plot(x, y, color)


def plot_lines(x, y, color):
    global ax
    ax.plot(x, y, color)


def pyplot_setup(xMin, xMax, yMin, yMax):
    global ax
    fig = plt.figure()
    ax = fig.add_subplot(111)

    xrange = [xMin, xMax]
    yrange = [yMin, yMax]
    ax.set_xlim(*xrange)
    ax.set_ylim(*yrange)
    ax.set_aspect(1)


# Returns the length of a file
def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
        f.close()
    return i + 1


def calculate_intersection(pointK, pointL, pointM, pointN):
    pointS=[]
    pointS.append(((pointN[0]-pointM[0])*(pointL[0]*pointK[1]-pointK[0]*pointL[1])-(pointL[0]-pointK[0])*(pointN[0]*pointM[1]-pointM[0]*pointN[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    pointS.append(((pointK[1]-pointL[1])*(pointN[0]*pointM[1]-pointM[0]*pointN[1])-(pointM[1]-pointN[1])*(pointL[0]*pointK[1]-pointK[0]*pointL[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    return pointS






#filename_input = raw_input("Gerber-File as input(default Test Blinker-F.Cu.gtl)") or ("Blinker-F.Cu.gtl")
filename_input = raw_input("Gerber-File as input(default Test footprints-F.Cu.gbr)") or ("Test footprints-F.Cu.gbr")
#filename_input = raw_input("Gerber-File as input(default Test LED-Driver V1-B.Cu.gbr)") or ("LED-Driver V1-B.Cu.gbr")
laser_diameter = float(raw_input("Assumed laserdiameter in mm (default 0.2)") or ("0.2"))
if (raw_input("Set xMin and yMin to zero (default Y)") or ("Y")) == "Y":
    set_xMin_yMin_to_zero = True
else:
    set_xMin_yMin_to_zero = False
border = float(raw_input("Border around your layout in mm (default 5)") or ("5"))
distance_fillinglines = float(raw_input("Distance between filling lines im mm (default 0.15)") or ("0.15"))
speed_perimeters = int(raw_input("Perimeterspeed in mm/min (default 1000)") or ("1000"))
speed_filling = int(raw_input("Fillingspeed in mm/min (default 1600)") or ("1600"))
#optimizing_cycles = int(raw_input("Number of optimizing cycles to reduce positioning (default 20)") or ("20"))
filename_output = raw_input("ngc-File as output (default output.ngc)") or ("output.ngc")


import_as_perimeter = False
geometry = gbr_import(filename_input)
geometry =  cascaded_union(geometry).boundary

import_as_perimeter = True
perimeter = gbr_import(filename_input)
perimeter =  cascaded_union(perimeter).boundary

xMin = int(geometry.bounds[0] - border)
yMin = int(geometry.bounds[1] - border)
xMax = int(geometry.bounds[2] + border)
yMax = int(geometry.bounds[3] + border)

if set_xMin_yMin_to_zero:
    geometry = translate(geometry, xMin * (-1), yMin * (-1))
    perimeter = translate(perimeter, xMin * (-1), yMin * (-1))
    xMax = xMax - xMin
    xMin = 0
    yMax = yMax - yMin
    yMin = 0

pyplot_setup(xMin, xMax, yMin, yMax)

print("")
print("Plotting geometry")
plot_geometry(geometry, 'black')


perimeter = optimizing_outline_arrangement_V2(perimeter)
#perimeter = optimizing_outline_arrangement(perimeter)

print("")
print("Plotting positioning")
plot_positioning(perimeter, 'green')
    
print("")
print("Plotting perimeter")
plot_geometry(perimeter, 'red')

filling_pattern = calculate_filling_pattern(eliminate_vertical_lines(explode_to_lines(perimeter)))

print("")
print("Plotting fillingpattern")
plot_fillingpattern(filling_pattern, 'red')

export_gcode(perimeter, filling_pattern)

plt.show()
